/* Phase 3 yacc file
   Nick Toothman
   CS 450

   Changes to BNF reflect the less stringent requirements for phase 3
   -no nested blocks
   -limitations on logic and control
   -context-sensitive components emulated

*/

%{

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "symtab.h"
#include "token.h"
#include "globals.h"

#define YYDEBUG 1

#define true 1
#define false 0

extern char tokenString[MAXTOKENLEN+1];   // from scanner 
extern int line_no;
extern int yyleng;
extern int col_no;

static Token* tokenList[MAXTOKENLEN];
static int listSize = 0;

static int address = 0;
static int params = 0;

static char cName[MAXTOKENLEN+1];
static char fName[MAXTOKENLEN+1];
static char stringTemp[MAXTOKENLEN+1];

void errorMsg(const char* errorMessage);

%}

%union {
 char c;
 int i;
 struct Token *t;
}


%token <i> IF THEN ELSE END RETURN UNTIL READ WRITE CLASS
%token <i> ASSIGN PLUS MINUS TIMES OVER LPAREN RPAREN SEMI COMMA LBRACE RBRACE
%token <i> ENDFILE ERROR 
%token <c> CHARLIT

%token <t> ID INT REAL 

%type <t> program fdec stmts  stmt  params factor term exp id num list_item list_seq
%type <t> flist fID fcall cID


%% /* cflat BNF grammar */

program     : CLASS cID LBRACE flist RBRACE 
            { 
              $$ = $2; 
	    }
            ;

cID    : ID
            {
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = strdup(tokenString);
              printf("class %s: \n", $$->name);
              strcpy(cName, tokenString);
            }
            ;
flist   : flist fdec
            {
              // empty body to shut up bison 
            }
            | fdec
            ;

fdec    : fID LPAREN params RPAREN LBRACE stmts RBRACE
            { 
              $$ = $1; 
            }
            ;


fID     : ID
            {
	      // create new token for function id, store in
	      // symbol table
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = strdup(tokenString);
              $$->type = 'F'; 
              $$->data = '-';
              $$->rValue = "void";
              $$->params = 0;
              printf("function %s from class %s:\n", $$->name, cName);
              strcpy(fName, tokenString);

              insert($$->name, yyleng, line_no, $$);
            }
            ;
params   : params id
            // increments number of parameters for the current function
	    {
              Token *f = getToken(fName);
              f->params++;
            }
            | id
            {
              Token *f = getToken(fName);
              f->params++;
            }
            | /*nothing*/ {}
            ;

stmts    : stmts stmt SEMI 
	    // reset listSize to 0 for 3 add code generation
            { listSize = 0; }
            | stmt SEMI
            { listSize = 0; }
            ;

stmt        : id ASSIGN exp
            { 
              $$ = $1; 

              // check type
              if ($1->data == 'V')
              {
                $1->data = tokenList[0]->data;
              }

              // gen 3 add code
              int i = 0;
              int j = 0;
              int found = false;
              char temp[MAXTOKENLEN+1];
	      while (listSize > 0)
              {
                switch (listSize)
                {
                  case 1:
                    printf("  %s = %s \n", $1->name, tokenList[0]->name);
                    if ($1->data != tokenList[0]->data)
                      errorMsg("type mismatch");
                    listSize -= 1;
                    break;

                  case 3:
                    printf("  %s = %s %s %s \n",
                      $1->name, tokenList[0]->name, tokenList[2]->name, tokenList[1]->name);
                    if ($1->data != tokenList[0]->data || $1->data != tokenList[1]->data)
                      errorMsg("type mismatch");
                    listSize -= 3;
                    break;

                  default:
                    // look for *
                    found = false;
                    for (i = 0; i < listSize && !found; i++)
                    {
                      if (tokenList[i]->name == "*") 
                        found = true;
                    }
                    if (found)
                    {
                      i--; // correct off by one
                      // print 3 add code
                      sprintf(temp, "t%02d", address++);
                      printf("  %s = %s * %s \n",
                        temp, tokenList[i-2]->name, tokenList[i-1]->name); 
                      if($1->data != tokenList[i-2]->data || $1->data != tokenList[i-1]->data)
                        errorMsg("type mismatch");

                      // replace with temp
                      tokenList[i-2] = (Token *) malloc(sizeof(Token));
                      tokenList[i-2]->name = strdup(temp);
                      tokenList[i-2]->type = 'T';
                      tokenList[i-2]->data = $1->data;

                      // move everything over
                      for (j = i + 1; j < listSize ; j++)
                        tokenList[j-2] = tokenList[j];
                      listSize -= 2;
                    }
                    else
                    {
                      // print 3 add code
                      sprintf(temp, "t%02d", address++);
                      printf("  %s = %s + %s \n",
                        temp, tokenList[0]->name, tokenList[1]->name);
                      if($1->data != tokenList[0]->data || $1->data != tokenList[1]->data)
                        errorMsg("type mismatch");

                      // replace with temp
                      tokenList[0] = (Token *) malloc(sizeof(Token));
                      tokenList[0]->name = strdup(temp);
                      tokenList[0]->type = 'T';
                      tokenList[0]->data = $1->data;

                      // move everything over
                      for(j = 3; j < listSize; j++)
                        tokenList[j-2] = tokenList[j];
                      listSize -= 2;
                    }
                }
              }
              

              // clean up 
              address = 0;
            }
            | fcall LPAREN list_seq RPAREN
            {
	      // look up function to get number of parameters for size comparison
              Token *f = getToken($1->name);
              if (f->params != params)
                errorMsg("incorrect number of parameters");
              printf("  function call: %s (%d parameter(s) found - ok) \n", $1->name, params);
              printf("  return type: %s\n", f->rValue);


              // reset current number of parameters to 0 for next function call
              params = 0;
            }
            | RETURN factor
            {
              Token *f = getToken(fName);
              f->rValue = $2->name;
            }
            ;
fcall   : ID
            {
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = strdup(tokenString);

              insert($$->name, strlen($$->name), line_no, $$);
            }
            ;
list_seq    : list_seq COMMA list_item
            | list_item
            | { /* shut bison up */ } 
            ;
list_item   : factor
            {
              printf("  param %s \n", $1->name);
              if ($1->data == 'V')
                errorMsg("parameter is void");
              params++;
            }
exp         : exp PLUS term
            {
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = "+";
              $$->type = 'O';
              $$->data = 'V';
              tokenList[listSize++] = $$;  // add term to tokenList and inc listSize for 3 add generation
            } 
            | term
            ;
term        : term TIMES factor
            {
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = "*";
              $$->type = 'O';
              $$->data = 'V';
              tokenList[listSize++] = $$;
            } 
            | factor
            ;
factor      : num
            {
              tokenList[listSize++] = $$;
            }
            | id
            {
              tokenList[listSize++] = $$;
            }
            ;
id          : ID
            {
	      // give the current ID a name in the symbol table
	      // that properly references its scope
              strcpy(stringTemp, fName);
              strcat(stringTemp, "_");
              strcat(stringTemp, tokenString);

              if (lookup(stringTemp))
              {
                $$ = getToken(stringTemp);
              }
              else
              {
                $$ = (Token *) malloc(sizeof(Token));
                $$->name = strdup(stringTemp);
                $$->type = 'I'; 
                $$->data = 'V';
              }

              insert($$->name, strlen($$->name), line_no, $$);
            }
            ;
num         : INT
            {
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = strdup(tokenString);
              $$->type = 'L';
              $$->data = 'I';

              insert($$->name, yyleng, line_no, $$);
            }
            | REAL
            {
              $$ = (Token *) malloc(sizeof(Token));
              $$->name = strdup(tokenString);
              $$->type = 'L';
              $$->data = 'R';

              insert($$->name, yyleng, line_no, $$);
            }
%%


void printToken( TokenType token, const char* tokenString );

extern FILE *yyin;
extern FILE *yyout;
extern int yyparse();
int yyerror(char * message);

int yydebug = 1;
extern int line_no;
extern char tokenString[MAXTOKENLEN+1];

int main( int argc,char *argv[] )
{
  if(argc < 2) {
     printf("Usage: cf <source filename>\n");
     exit(0);
  }
  FILE *fp = fopen(argv[1],"r");
  if(!fp) {
     printf("Unable to open file for reading\n");
     exit(0);
  }
  yyin = fp;

  fp = fopen("sym.output","w");
  if(!fp) {
     printf("Unable to open file for writing\n");
     exit(0);
  }

  int flag = yyparse();

/* cleanup */
  dump_symtab(stdout);   // dump symbol table 
  dump_mem();        // dump and free memory

  fclose(fp);
  fclose(yyin);


  return flag;
}


int yywrap()
{
  return(1);
}

int yyerror(char * message)
{ fprintf(stdout,"syntax error at line %d: %s\n",line_no,message);
  fprintf(stdout,"current token: ");
  printToken(yychar,tokenString);
  return 0;
}

void errorMsg(const char* errorMessage)
{
  printf("*** Error: %s \n", errorMessage);
}

/* useful for tracing parse */ 
void printToken( TokenType token, const char* tokenString )
{ switch (token)
  { case IF:
    case THEN:
    case ELSE:
    case END:
    case RETURN:
    case UNTIL:
    case READ:
    case WRITE:
      fprintf(stdout, "reserved word: %s\n",tokenString);
      break;
    case RPAREN: fprintf(stdout,")\n"); break;
    case LPAREN: fprintf(stdout,"(\n"); break;
    case TIMES: fprintf(stdout,"*\n"); break;
    case PLUS: fprintf(stdout,"+\n"); break;
    case INT:
      fprintf(stdout, "NUM, val= %s\n",tokenString);
      break;
    case REAL:
      fprintf(stdout, "NUM, val= %s\n",tokenString);
      break;
    case ID:
      fprintf(stdout, "ID, name= %s\n",tokenString);
      break;
    case ERROR:
fprintf(stdout, "ERROR: %s\n",tokenString);
      break;
    default: 
      fprintf(stdout,"Unknown token: %s\n",tokenString);
  }
}




